// Copyright (C) 2006-2014 David Sugar, Tycho Softworks.
// Copyright (C) 2015-2020 Cherokees of Idaho.
//
// This file is part of GNU uCommon C++.
//
// GNU uCommon C++ is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GNU uCommon C++ is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GNU uCommon C++.  If not, see <http://www.gnu.org/licenses/>.

/**
 * Placeholder for future classes that require stl support.
 * @file ucommon/stl.h
 */

#ifndef	UCOMMON_SYSRUNTIME
#ifndef _UCOMMON_STL_H_
#define _UCOMMON_STL_H_

#ifndef _UCOMMON_PLATFORM_H_
#include <ucommon/platform.h>
#endif

#include <list> // example...

namespace ucommon {

/*
    In the future we may introduce optional classes which require and/or
    build upon the standard template library.  This header indicates how and
    where they may be added.
*/

} // namespace ucommon

#endif
#endif
